from nebullvm.apps.base import App

from speedster.root_op import SpeedsterRootOp


class SpeedsterApp(App):
    def __init__(self):
        super().__init__()
        self.root_op = SpeedsterRootOp()

    def execute(self, *args, **kwargs):
        return self.root_op.execute(*args, **kwargs)



from nebullvm.config import COMPILER_LIST, COMPRESSOR_LIST
from nebullvm.operations.inference_learners.huggingface import (
    HuggingFaceInferenceLearner,
)
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch
from transformers import AlbertModel, TFAlbertModel, AlbertTokenizer

from speedster import optimize_model


def test_torch_huggingface_ort_input_text():
    tokenizer = AlbertTokenizer.from_pretrained("albert-base-v1")
    model = AlbertModel.from_pretrained("albert-base-v1")

    # Move the model to gpu if available
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    model.eval()

    input_data = [
        "this is a test",
        "hi my name is Valerio",
        "india is very far from italy",
    ]

    optimized_model = optimize_model(
        model=model,
        input_data=input_data,
        optimization_time="constrained",
        tokenizer=tokenizer,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "onnxruntime"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
        tokenizer_args=dict(
            add_special_tokens=True,
            return_attention_mask=True,
            return_tensors="pt",
            return_token_type_ids=None,  # Sets to model default
            padding="longest",
            truncation=True,
        ),
    )

    x = ["this is a test input to see if the optimized model works."]
    inputs = tokenizer(x, return_tensors="pt").to(device)
    res_original = model(**inputs)
    res_optimized = optimized_model(**inputs)

    assert isinstance(optimized_model, HuggingFaceInferenceLearner)

    assert (
        torch.max(
            abs(
                (
                    res_original["last_hidden_state"]
                    - res_optimized["last_hidden_state"]
                )
            )
        )
        < 1e-2
    )
    assert (
        torch.max(
            abs(
                (
                    res_original["pooler_output"]
                    - res_optimized["pooler_output"]
                )
            )
        )
        < 1e-2
    )


def test_torch_huggingface_ort_input_tensors():
    tokenizer = AlbertTokenizer.from_pretrained("albert-base-v1")
    model = AlbertModel.from_pretrained("albert-base-v1")

    # Move the model to gpu if available
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    model.eval()

    text = "hi my name is Valerio"
    inputs = tokenizer(text, return_tensors="pt").to(device)

    dynamic_info = {
        "inputs": [
            {0: "batch", 1: "num_tokens"},
            {0: "batch", 1: "num_tokens"},
            {0: "batch", 1: "num_tokens"},
        ],
        "outputs": [{0: "batch", 1: "num_tokens"}, {0: "batch"}],
    }

    optimized_model = optimize_model(
        model=model,
        input_data=[inputs for _ in range(10)],
        optimization_time="constrained",
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "onnxruntime"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
        dynamic_info=dynamic_info,
    )

    x = ["this is a test input to see if the optimized model works."]
    inputs = tokenizer(x, return_tensors="pt").to(device)
    res_original = model(**inputs)
    res_optimized = optimized_model(**inputs)

    assert isinstance(optimized_model, HuggingFaceInferenceLearner)

    assert (
        torch.max(
            abs(
                (
                    res_original["last_hidden_state"]
                    - res_optimized["last_hidden_state"]
                )
            )
        )
        < 1e-2
    )
    assert (
        torch.max(
            abs(
                (
                    res_original["pooler_output"]
                    - res_optimized["pooler_output"]
                )
            )
        )
        < 1e-2
    )


def test_torch_huggingface_torchscript_input_tensors():
    tokenizer = AlbertTokenizer.from_pretrained("albert-base-v1")
    model = AlbertModel.from_pretrained("albert-base-v1", torchscript=True)

    # Move the model to gpu if available
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    model.eval()

    text = "hi my name is Valerio"
    inputs = tokenizer(text, return_tensors="pt").to(device)

    dynamic_info = {
        "inputs": [
            {0: "batch", 1: "num_tokens"},
            {0: "batch", 1: "num_tokens"},
            {0: "batch", 1: "num_tokens"},
        ],
        "outputs": [{0: "batch", 1: "num_tokens"}, {0: "batch"}],
    }

    optimized_model = optimize_model(
        model=model,
        input_data=[inputs for _ in range(10)],
        optimization_time="constrained",
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "torchscript"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
        dynamic_info=dynamic_info,
    )

    x = ["this is a test input to see if the optimized model works."]
    inputs = tokenizer(x, return_tensors="pt").to(device)
    res_original = model(**inputs)
    res_optimized = optimized_model(**inputs)

    assert isinstance(optimized_model, HuggingFaceInferenceLearner)

    assert torch.max(abs((res_original[0] - res_optimized[0]))) < 1e-2
    assert torch.max(abs((res_original[1] - res_optimized[1]))) < 1e-2


def test_tensorflow_huggingface_ort_input_text_np():
    tokenizer = AlbertTokenizer.from_pretrained("albert-base-v1")
    model = TFAlbertModel.from_pretrained("albert-base-v1")

    input_data = [
        "this is a test",
        "hi my name is Valerio",
        "india is very far from italy",
    ]

    dynamic_info = {
        "inputs": [
            {0: "batch", 1: "num_tokens"},
            {0: "batch", 1: "num_tokens"},
            {0: "batch", 1: "num_tokens"},
        ],
        "outputs": [{0: "batch", 1: "num_tokens"}, {0: "batch"}],
    }

    optimized_model = optimize_model(
        model=model,
        input_data=input_data,
        optimization_time="constrained",
        tokenizer=tokenizer,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "onnxruntime"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
        tokenizer_args=dict(
            add_special_tokens=True,
            return_attention_mask=True,
            return_tensors="np",
            return_token_type_ids=None,  # Sets to model default
            padding="longest",
            truncation=True,
        ),
        dynamic_info=dynamic_info,
    )

    x = ["this is a test input to see if the optimized model works."]
    inputs = tokenizer(x, return_tensors="np")
    res_original = model(**inputs)
    res_optimized = optimized_model(**inputs)

    assert isinstance(optimized_model, HuggingFaceInferenceLearner)

    assert (
        tf.math.reduce_max(
            abs(
                (
                    res_original["last_hidden_state"]
                    - res_optimized["last_hidden_state"]
                )
            )
        )
        < 1e-2
    )
    assert (
        tf.math.reduce_max(
            abs(
                (
                    res_original["pooler_output"]
                    - res_optimized["pooler_output"]
                )
            )
        )
        < 1e-2
    )


def test_tensorflow_huggingface_ort_input_tensors_np():
    tokenizer = AlbertTokenizer.from_pretrained("albert-base-v1")
    model = TFAlbertModel.from_pretrained("albert-base-v1")

    text = "hi my name is Valerio"
    inputs = tokenizer(text, return_tensors="np")

    dynamic_info = {
        "inputs": [
            {0: "batch", 1: "num_tokens"},
            {0: "batch", 1: "num_tokens"},
            {0: "batch", 1: "num_tokens"},
        ],
        "outputs": [{0: "batch", 1: "num_tokens"}, {0: "batch"}],
    }

    optimized_model = optimize_model(
        model=model,
        input_data=[inputs for _ in range(10)],
        optimization_time="constrained",
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "onnxruntime"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
        dynamic_info=dynamic_info,
    )

    x = ["Test to see if it works with a different output"]
    inputs = tokenizer(x, return_tensors="np")
    res_original = model(**inputs)
    res_optimized = optimized_model(**inputs)

    assert isinstance(optimized_model, HuggingFaceInferenceLearner)

    assert (
        tf.math.reduce_max(
            abs(
                (
                    res_original["last_hidden_state"]
                    - res_optimized["last_hidden_state"]
                )
            )
        )
        < 1e-2
    )
    assert (
        tf.math.reduce_max(
            abs(
                (
                    res_original["pooler_output"]
                    - res_optimized["pooler_output"]
                )
            )
        )
        < 1e-2
    )


def test_tensorflow_huggingface_ort_input_text_tf():
    tokenizer = AlbertTokenizer.from_pretrained("albert-base-v1")
    model = TFAlbertModel.from_pretrained("albert-base-v1")

    input_data = [
        "this is a test",
        "hi my name is Valerio",
        "india is very far from italy",
    ]

    dynamic_info = {
        "inputs": [
            {0: "batch", 1: "num_tokens"},
            {0: "batch", 1: "num_tokens"},
            {0: "batch", 1: "num_tokens"},
        ],
        "outputs": [{0: "batch", 1: "num_tokens"}, {0: "batch"}],
    }

    optimized_model = optimize_model(
        model=model,
        input_data=input_data,
        optimization_time="constrained",
        tokenizer=tokenizer,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "onnxruntime"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
        tokenizer_args=dict(
            add_special_tokens=True,
            return_attention_mask=True,
            return_tensors="tf",
            return_token_type_ids=None,  # Sets to model default
            padding="longest",
            truncation=True,
        ),
        dynamic_info=dynamic_info,
    )

    x = ["this is a test input to see if the optimized model works."]
    inputs = tokenizer(x, return_tensors="tf")
    res_original = model(**inputs)
    res_optimized = optimized_model(**inputs)

    assert isinstance(optimized_model, HuggingFaceInferenceLearner)

    assert (
        tf.math.reduce_max(
            abs(
                (
                    res_original["last_hidden_state"]
                    - res_optimized["last_hidden_state"]
                )
            )
        )
        < 1e-2
    )
    assert (
        tf.math.reduce_max(
            abs(
                (
                    res_original["pooler_output"]
                    - res_optimized["pooler_output"]
                )
            )
        )
        < 1e-2
    )


def test_tensorflow_huggingface_ort_input_tensors_tf():
    tokenizer = AlbertTokenizer.from_pretrained("albert-base-v1")
    model = TFAlbertModel.from_pretrained("albert-base-v1")

    text = "hi my name is Valerio"
    inputs = tokenizer(text, return_tensors="tf")

    dynamic_info = {
        "inputs": [
            {0: "batch", 1: "num_tokens"},
            {0: "batch", 1: "num_tokens"},
            {0: "batch", 1: "num_tokens"},
        ],
        "outputs": [{0: "batch", 1: "num_tokens"}, {0: "batch"}],
    }

    optimized_model = optimize_model(
        model=model,
        input_data=[inputs for _ in range(10)],
        optimization_time="constrained",
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "onnxruntime"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
        dynamic_info=dynamic_info,
    )

    x = ["Test to see if it works with a different output"]
    inputs = tokenizer(x, return_tensors="tf")
    res_original = model(**inputs)
    res_optimized = optimized_model(**inputs)

    assert isinstance(optimized_model, HuggingFaceInferenceLearner)

    assert (
        tf.math.reduce_max(
            abs(
                (
                    res_original["last_hidden_state"]
                    - res_optimized["last_hidden_state"]
                )
            )
        )
        < 1e-2
    )
    assert (
        tf.math.reduce_max(
            abs(
                (
                    res_original["pooler_output"]
                    - res_optimized["pooler_output"]
                )
            )
        )
        < 1e-2
    )



import os
import platform
import uuid
from pathlib import Path
from typing import Any

import cpuinfo
import psutil

from nebullvm.optional_modules.torch import Module
from nebullvm.optional_modules.utils import (
    torch_is_available,
    tensorflow_is_available,
)
from nebullvm.tools.base import Device
from nebullvm.tools.pytorch import torch_get_device_name
from nebullvm.tools.tf import tensorflow_get_gpu_name


def generate_model_id(model_name: str):
    return f"{str(uuid.uuid4())}_{hash(model_name)}"


def get_model_name(model: Any):
    if isinstance(model, str) or isinstance(model, Path):
        return str(model)
    else:
        return model.__class__.__name__


def _get_gpu_name():
    if torch_is_available():
        name = torch_get_device_name()
    elif tensorflow_is_available():
        name = tensorflow_get_gpu_name()
    else:
        name = "Unknown GPU"

    return name


def get_hw_info(device: Device):
    hw_info = {
        "cpu": cpuinfo.get_cpu_info()["brand_raw"],
        "operative_system": platform.system(),
        "ram": f"{round(psutil.virtual_memory().total * 1e-9, 2)} GB",
    }
    if device is Device.GPU:
        hw_info["gpu"] = _get_gpu_name()
    return hw_info


def read_model_size(model: Any):
    if isinstance(model, str) or isinstance(model, Path):
        size = os.stat(str(model)).st_size
    elif isinstance(model, Module):
        size = sum(
            param.nelement() * param.element_size()
            for param in model.parameters()
        )
    else:
        # we assume it is a tf_model
        size = model.count_params() * 4  # assuming full precision 32 bit
    return f"{round(size * 1e-6, 2)} MB"



import os
from pathlib import Path

from nebullvm.operations.conversions.pytorch import convert_torch_to_onnx
from nebullvm.tools.base import ModelParams, Device
from nebullvm.tools.data import DataManager
from nebullvm.tools.utils import gpu_is_available


def torch_to_onnx(model, input_data, output_path):
    model_params = ModelParams(1, [], [])
    output_path = os.path.join(output_path, "model.onnx")
    device = Device.GPU if gpu_is_available() else Device.CPU
    convert_torch_to_onnx(
        model, DataManager(input_data), model_params, Path(output_path), device
    )

    return output_path



import cpuinfo

import pytest
import tensorflow as tf
from keras.applications import ResNet50
from nebullvm.config import COMPILER_LIST, COMPRESSOR_LIST
from nebullvm.operations.inference_learners.onnx import (
    TensorflowONNXInferenceLearner,
)
from nebullvm.operations.inference_learners.openvino import (
    TensorflowOpenVinoInferenceLearner,
)
from nebullvm.operations.inference_learners.tensor_rt import (
    TensorflowONNXTensorRTInferenceLearner,
)
from nebullvm.operations.inference_learners.tensorflow import (
    TensorflowBackendInferenceLearner,
    TFLiteBackendInferenceLearner,
)
from nebullvm.operations.inference_learners.tvm import (
    TensorflowApacheTVMInferenceLearner,
)
from nebullvm.operations.optimizations.compilers.utils import tvm_is_available
from nebullvm.tools.utils import gpu_is_available

from speedster import optimize_model

# Limit tensorflow gpu memory usage
gpus = tf.config.list_physical_devices("GPU")
if gpus:
    try:
        # Currently, memory growth needs to be the same across GPUs
        for gpu in gpus:
            tf.config.set_visible_devices(gpus[0], "GPU")
            tf.config.experimental.set_memory_growth(gpu, True)
            logical_gpus = tf.config.list_logical_devices("GPU")
            print(
                len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs"
            )
    except RuntimeError as e:
        # Memory growth must be set before GPUs have been initialized
        print(e)


def test_tensorflow_ort():
    model = ResNet50()
    input_data = [
        ((tf.random.normal([1, 224, 224, 3]),), 0) for i in range(100)
    ]

    # Run nebullvm optimization in one line of code
    optimized_model = optimize_model(
        model,
        input_data=input_data,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "onnxruntime"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
    )

    # Try the optimized model
    x = tf.random.normal([1, 224, 224, 3])
    res_original = model.predict(x)
    res_optimized = optimized_model.predict(x)[0]

    assert isinstance(optimized_model, TensorflowONNXInferenceLearner)
    assert abs((res_original - res_optimized)).max() < 1e-2


def test_tensorflow_tf_backend():
    model = ResNet50()
    input_data = [
        ((tf.random.normal([1, 224, 224, 3]),), 0) for i in range(100)
    ]

    # Run nebullvm optimization in one line of code
    optimized_model = optimize_model(
        model,
        input_data=input_data,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "xla"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
    )

    # Try the optimized model
    x = tf.random.normal([1, 224, 224, 3])
    res_original = model.predict(x)
    res_optimized = optimized_model.predict(x)[0]

    assert isinstance(optimized_model, TensorflowBackendInferenceLearner)
    assert abs((res_original - res_optimized)).max() < 1e-2


@pytest.mark.skipif(
    gpu_is_available(),
    reason="TFLite does not support Nvidia GPUs",
)
def test_tensorflow_tflite():
    model = ResNet50()
    input_data = [
        ((tf.random.normal([1, 224, 224, 3]),), 0) for i in range(100)
    ]

    # Run nebullvm optimization in one line of code
    optimized_model = optimize_model(
        model,
        input_data=input_data,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "tflite"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
        metric_drop_ths=0.1,
    )

    # Try the optimized model
    x = tf.random.normal([1, 224, 224, 3])
    res_original = model.predict(x)
    res_optimized = optimized_model.predict(x)[0]

    assert isinstance(optimized_model, TFLiteBackendInferenceLearner)
    assert abs((res_original - res_optimized)).max() < 1e-2


@pytest.mark.skipif(
    not gpu_is_available(),
    reason="Skip because cuda is not available.",
)
def test_tensorflow_tensorrt():
    model = ResNet50()
    input_data = [
        ((tf.random.normal([1, 224, 224, 3]),), 0) for i in range(100)
    ]

    # Run nebullvm optimization in one line of code
    optimized_model = optimize_model(
        model,
        input_data=input_data,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "tensor_rt"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
    )

    # Try the optimized model
    x = tf.random.normal([1, 224, 224, 3])
    res_original = model.predict(x)
    res_optimized = optimized_model.predict(x)[0]

    assert isinstance(optimized_model, TensorflowONNXTensorRTInferenceLearner)
    assert abs((res_original - res_optimized)).max() < 1e-2


@pytest.mark.skipif(
    "intel" not in cpuinfo.get_cpu_info()["brand_raw"].lower(),
    reason="Openvino is only available for intel processors.",
)
def test_tensorflow_openvino():
    model = ResNet50()
    input_data = [
        ((tf.random.normal([1, 224, 224, 3]),), 0) for i in range(100)
    ]

    # Run nebullvm optimization in one line of code
    optimized_model = optimize_model(
        model,
        input_data=input_data,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "openvino"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
        device="cpu",
    )

    # Try the optimized model
    x = tf.random.normal([1, 224, 224, 3])
    res_original = model.predict(x)
    res_optimized = optimized_model.predict(x)[0]

    assert isinstance(optimized_model, TensorflowOpenVinoInferenceLearner)
    assert abs((res_original - res_optimized)).max() < 1e-2


@pytest.mark.skipif(
    not tvm_is_available(), reason="Can't test tvm if it's not installed."
)
def test_tensorflow_tvm():
    model = ResNet50()
    input_data = [
        ((tf.random.normal([1, 224, 224, 3]),), 0) for i in range(100)
    ]

    # Run nebullvm optimization in one line of code
    optimized_model = optimize_model(
        model,
        input_data=input_data,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "tvm"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
    )

    # Try the optimized model
    x = tf.random.normal([1, 224, 224, 3])
    res_original = model.predict(x)
    res_optimized = optimized_model.predict(x)[0]

    assert isinstance(optimized_model, TensorflowApacheTVMInferenceLearner)
    assert abs((res_original - res_optimized)).max() < 1e-2



import cpuinfo

import pytest
import torch
import torchvision.models as models
from nebullvm.config import COMPILER_LIST, COMPRESSOR_LIST
from nebullvm.operations.inference_learners.blade_disc import (
    BladeDISCInferenceLearner,
)
from nebullvm.operations.inference_learners.onnx import (
    PytorchONNXInferenceLearner,
)
from nebullvm.operations.inference_learners.openvino import (
    PytorchOpenVinoInferenceLearner,
)
from nebullvm.operations.inference_learners.pytorch import (
    PytorchBackendInferenceLearner,
)
from nebullvm.operations.inference_learners.tensor_rt import (
    PytorchTensorRTInferenceLearner,
    PytorchONNXTensorRTInferenceLearner,
)
from nebullvm.operations.inference_learners.tvm import (
    PytorchApacheTVMInferenceLearner,
)
from nebullvm.operations.optimizations.compilers.utils import (
    tvm_is_available,
    bladedisc_is_available,
)

from speedster import optimize_model


def test_torch_ort():
    model = models.resnet18()
    input_data = [((torch.randn(1, 3, 256, 256),), 0) for i in range(100)]

    # Run nebullvm optimization in one line of code
    optimized_model = optimize_model(
        model,
        input_data=input_data,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "onnxruntime"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
    )

    # Try the optimized model
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    x = torch.randn(1, 3, 256, 256, requires_grad=False).to(device)
    model.eval()
    res_original = model(x)
    res_optimized = optimized_model(x)[0]

    assert isinstance(optimized_model, PytorchONNXInferenceLearner)
    assert torch.max(abs((res_original - res_optimized))) < 1e-2


def test_torch_ort_quant():
    model = models.resnet18()
    input_data = [((torch.randn(1, 3, 256, 256),), 0) for i in range(100)]

    # Run nebullvm optimization in one line of code
    optimized_model = optimize_model(
        model,
        input_data=input_data,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "onnxruntime"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
        metric_drop_ths=2,
    )

    # Try the optimized model
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    x = torch.randn(1, 3, 256, 256, requires_grad=False).to(device)
    model.eval()
    res_original = model(x)
    res_optimized = optimized_model(x)[0]

    assert isinstance(optimized_model, PytorchONNXInferenceLearner)
    assert torch.max(abs((res_original - res_optimized))) < 2


def test_torch_torchscript():
    model = models.resnet18()
    input_data = [((torch.randn(1, 3, 256, 256),), 0) for i in range(100)]

    # Run nebullvm optimization in one line of code
    optimized_model = optimize_model(
        model,
        input_data=input_data,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "torchscript"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
    )

    # Try the optimized model
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    x = torch.randn(1, 3, 256, 256, requires_grad=False).to(device)
    model.eval()
    res_original = model(x)
    res_optimized = optimized_model(x)[0]

    assert isinstance(optimized_model, PytorchBackendInferenceLearner)
    assert torch.max(abs((res_original - res_optimized))) < 1e-2


@pytest.mark.skipif(
    not torch.cuda.is_available(),
    reason="Skip because cuda is not available.",
)
def test_torch_tensorrt():
    model = models.resnet18()
    input_data = [((torch.randn(1, 3, 256, 256),), 0) for i in range(100)]

    # Run nebullvm optimization in one line of code
    optimized_model = optimize_model(
        model,
        input_data=input_data,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "tensor_rt"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
    )

    # Try the optimized model
    x = torch.randn(1, 3, 256, 256).cuda()
    model.eval()
    res_original = model(x)
    res_optimized = optimized_model(x)[0]

    assert isinstance(
        optimized_model, PytorchTensorRTInferenceLearner
    ) or isinstance(optimized_model, PytorchONNXTensorRTInferenceLearner)
    assert torch.max(abs((res_original - res_optimized))) < 1e-2


@pytest.mark.skipif(
    "intel" not in cpuinfo.get_cpu_info()["brand_raw"].lower(),
    reason="Openvino is only available for intel processors.",
)
def test_torch_openvino():
    model = models.resnet18()
    input_data = [((torch.randn(1, 3, 256, 256),), 0) for i in range(100)]

    # Run nebullvm optimization in one line of code
    optimized_model = optimize_model(
        model,
        input_data=input_data,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "openvino"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
        device="cpu",
    )

    # Try the optimized model
    x = torch.randn(1, 3, 256, 256)
    model.eval()
    res_original = model(x)
    res_optimized = optimized_model(x)[0]

    assert isinstance(optimized_model, PytorchOpenVinoInferenceLearner)
    assert torch.max(abs((res_original.cpu() - res_optimized))) < 1e-2


@pytest.mark.skipif(
    not tvm_is_available(), reason="Can't test tvm if it's not installed."
)
def test_torch_tvm():
    model = models.resnet18()
    input_data = [((torch.randn(1, 3, 256, 256),), 0) for i in range(100)]

    # Run nebullvm optimization in one line of code
    optimized_model = optimize_model(
        model,
        input_data=input_data,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "tvm"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
    )

    # Try the optimized model
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    x = torch.randn(1, 3, 256, 256, requires_grad=False).to(device)
    model.eval()
    res_original = model(x)
    res_optimized = optimized_model(x)[0]

    assert isinstance(optimized_model, PytorchApacheTVMInferenceLearner)
    assert torch.max(abs((res_original - res_optimized))) < 1e-2


@pytest.mark.skipif(
    not bladedisc_is_available(),
    reason="Can't test bladedisc if it's not installed.",
)
def test_torch_bladedisc():
    model = models.resnet18()
    input_data = [((torch.randn(1, 3, 256, 256),), 0) for i in range(100)]

    # Run nebullvm optimization in one line of code
    optimized_model = optimize_model(
        model,
        input_data=input_data,
        ignore_compilers=[
            compiler for compiler in COMPILER_LIST if compiler != "bladedisc"
        ],
        ignore_compressors=[compressor for compressor in COMPRESSOR_LIST],
    )

    # Try the optimized model
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    x = torch.randn(1, 3, 256, 256, requires_grad=False).to(device)
    model.eval()
    res_original = model(x)
    res_optimized = optimized_model(x)[0]

    assert isinstance(optimized_model, BladeDISCInferenceLearner)
    assert torch.max(abs((res_original - res_optimized))) < 1e-2



import logging
from typing import (
    Union,
    Iterable,
    Sequence,
    Callable,
    Dict,
    List,
    Optional,
)

from loguru import logger
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import torch
from nebullvm.tools.base import Device
from nebullvm.tools.logger import debug_mode_enabled, LoggingContext
from nebullvm.tools.utils import gpu_is_available

from speedster.root_op import SpeedsterRootOp


def _check_device(device: Optional[str]) -> Device:
    if device is None:
        if gpu_is_available():
            device = Device.GPU
        else:
            device = Device.CPU
    else:
        if device.lower() == "gpu":
            if not gpu_is_available():
                logger.warning(
                    "Selected GPU device but no available GPU found on this "
                    "platform. CPU will be used instead. Please make sure "
                    "that the gpu is installed and can be used by your "
                    "framework."
                )
                device = Device.CPU
            else:
                device = Device.GPU
        else:
            device = Device.CPU

    return device


def optimize_model(
    model: Union[torch.nn.Module, tf.Module, str],
    input_data: Union[Iterable, Sequence],
    metric_drop_ths: float = None,
    metric: Union[str, Callable] = None,
    optimization_time: str = "constrained",
    dynamic_info: Dict = None,
    config_file: str = None,
    ignore_compilers: List[str] = None,
    ignore_compressors: List[str] = None,
    store_latencies: bool = False,
    device: Optional[str] = None,
    **kwargs,
):
    """Optimize the input model regardless of the framework it was used for
    implementing it. The optimized model given as output will share with the
    input one the same API, i.e. the optimized model will have the same
    interface as the original one.

    Args:
        model (Union[torch.Module, tf.Module, str]): The input model. It can be
            a torch or tensorflow model or a path to an onnx saved model.
        input_data (Iterable or Sequence): Input data to be used for
            optimizing the model. Note that if 'unconstrained' is selected as
            `optimization_time`, it would be beneficial to provide at least 100
            data samples in order to use all the techniques supported by
            Nebullvm. The data can be given in either as sequence (data can be
            accessed by "element", e.g. `data[i]`) or iterable (data needs to
            be accessed with loop, e.g. `for x in data`). PyTorch, TensorFlow
            and Onnx respectively accept input tensor in `torch.Tensor`,
            `tf.Tensor` and `np.ndarray` formats. Note that each input
            sample must be a tuple containing a tuple as first element, the
            `inputs`, and the `label` as second element. The `inputs` needs to
            be passed as tuple even if a single input is needed by the model
            (in this case the `inputs` tuple will contain just an element).
            HuggingFace models can take as data samples both dictionaries or
            strings. Strings will then be converted in data samples using the
            HuggingFace tokenizer which must be given as input when just a
            list of string is provided as input_data (tokenizers can be passed
            as extra arguments of this function using the keyword `tokenizer`).
        metric_drop_ths (float, optional): Maximum reduction in the
            selected metric accepted. No model with a higher error will be
            accepted, i.e. all optimized model having a larger error respect to
            the original one will be discarded, without even considering their
            possible speed-up. Default: None, i.e. no drop in metric accepted.
        metric (Union[Callable, str], optional): The metric to
            be used for accepting or refusing a precision-reduction
            optimization proposal. If none is given but a `metric_drop_ths` is
            received, the `nebullvm.measure.compute_relative_difference`
            metric will be used as default one. A user-defined metric can
            be passed as function accepting as inputs two tuples of tensors
            (produced by the baseline and the optimized model) and the related
            original labels.
            For more information see
            `nebullvm.measure.compute_relative_difference` and
            `nebullvm.measure.compute_accuracy_drop`. `metric`
            accepts as value also a string containing the metric name. At the
            current stage the supported metrics are `"numeric_precision"` and
            `"accuracy"`. Default: `"numeric_precision"`
        optimization_time (OptimizationTime, optional): The optimization time
            mode. It can be either 'constrained' or 'unconstrained'. For
            'constrained' mode just compilers and precision reduction
            techniques are used (no compression). 'Unconstrained' optimization
            allows the usage of more time-consuming techniques as pruning and
            distillation. Note that for using many of the sophisticated
            techniques in the 'unconstrained' optimization, a small fine-tuning
            of the model will be needed. Thus we highly recommend to give as
            input_data at least 100 samples for when selecting 'unconstrained'
            optimization. Default: 'constrained'.
        dynamic_info (Dict, optional): Dictionary containing info about the
            dynamic axis. It should contain as keys both "inputs" and "outputs"
            and as values two lists of dictionaries where each dictionary
            represents the dynamic axis information for an input/output tensor.
            The inner dictionary should have as key an integer, i.e. the
            dynamic axis (considering also the batch size) and as value a
            string giving a "tag" to it, e.g. "batch_size". Default: None
        config_file (str, optional): Configuration file containing the
            parameters needed for defining the CompressionStep in the pipeline.
            Default: None.
        ignore_compilers (List, optional): List containing the compilers to be
            ignored during the OptimizerStep. The compiler name should be one
            among tvm, tensor RT, openvino, onnxruntime, deepsparse, tflite,
            bladedisc, torchscript, intel_neural_compressor. Default: None.
        ignore_compressors (List, optional): List containing the compressors
            to be ignored during the CompressionStep. The compiler name should
            be one among . Default: None.
        store_latencies (bool, optional): Parameter that allows to save the
            latency for each compiler used by nebullvm. Default: False.
        device (str, optional): Device used, can be 'cpu' or 'gpu'. If not
            set, gpu will be used if available, otherwise cpu. Default: None

    Returns:
        InferenceLearner: Optimized version of the input model having the same
            interface, imported by its original framework. For instance a
            Pytorch model, when optimized, will return an InferenceLearner
            object that can be call exactly as a PyTorch model (either
            with `model.forward(input)` and `model(input)`), i.e. it will
            take as input and it will return `torch.Tensor`s.
    """
    root_op = SpeedsterRootOp()
    device = _check_device(device)

    disable_log = True if not debug_mode_enabled() else False

    with LoggingContext(logging.getLogger(), disabled=disable_log):
        root_op.to(device).execute(
            model=model,
            input_data=input_data,
            metric_drop_ths=metric_drop_ths,
            metric=metric,
            optimization_time=optimization_time,
            dynamic_info=dynamic_info,
            config_file=config_file,
            ignore_compilers=ignore_compilers,
            ignore_compressors=ignore_compressors,
            store_latencies=store_latencies,
            **kwargs,
        )

    return root_op.get_result()



import json
import os
import pickle
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import (
    Any,
    Union,
    Iterable,
    Sequence,
    Dict,
    Callable,
    List,
)

from nebullvm.config import TRAIN_TEST_SPLIT_RATIO, MIN_NUMBER
from nebullvm.operations.base import Operation
from nebullvm.operations.conversions.converters import (
    PytorchConverter,
    TensorflowConverter,
    ONNXConverter,
)
from nebullvm.operations.conversions.huggingface import convert_hf_model
from nebullvm.operations.fetch_operations.local import (
    FetchModelFromLocal,
    FetchDataFromLocal,
)
from nebullvm.operations.inference_learners.base import BaseInferenceLearner
from nebullvm.operations.measures.measures import LatencyOriginalModelMeasure
from nebullvm.operations.measures.utils import QUANTIZATION_METRIC_MAP
from nebullvm.operations.optimizations.optimizers import (
    PytorchOptimizer,
    ONNXOptimizer,
    TensorflowOptimizer,
)
from nebullvm.operations.optimizations.utils import (
    map_compilers_and_compressors,
)
from nebullvm.optional_modules.tensorflow import tensorflow as tf
from nebullvm.optional_modules.torch import Module, DataLoader
from nebullvm.optional_modules.utils import check_dependencies
from nebullvm.tools.base import (
    ModelCompiler,
    DeepLearningFramework,
    ModelParams,
    OptimizationTime,
    ModelCompressor,
)
from nebullvm.tools.data import DataManager
from nebullvm.tools.feedback_collector import FeedbackCollector
from nebullvm.tools.utils import (
    get_dl_framework,
    is_huggingface_data,
    check_input_data,
    is_data_subscriptable,
    extract_info_from_data,
)
from tabulate import tabulate

from speedster.utils import (
    get_model_name,
    read_model_size,
    generate_model_id,
    get_hw_info,
)


SPEEDSTER_FEEDBACK_COLLECTOR = FeedbackCollector(
    url="https://nebuly.cloud/v1/store_speedster_results",
    disable_telemetry_environ_var="SPEEDSTER_DISABLE_TELEMETRY",
    app_version="0.1.0",
)


class SpeedsterRootOp(Operation):
    def __init__(self):
        super().__init__()

        self.model = None
        self.data = None
        self.optimal_model = None
        self.conversion_op = None

        self.fetch_model_op = FetchModelFromLocal()
        self.fetch_data_op = FetchDataFromLocal()
        self.orig_latency_measure_op = LatencyOriginalModelMeasure()
        self.torch_conversion_op = PytorchConverter()
        self.tensorflow_conversion_op = TensorflowConverter()
        self.torch_optimization_op = PytorchOptimizer()
        self.onnx_conversion_op = ONNXConverter()
        self.onnx_optimization_op = ONNXOptimizer()
        self.tensorflow_optimization_op = TensorflowOptimizer()
        self.set_feedback_collector(SPEEDSTER_FEEDBACK_COLLECTOR)

    def _get_conversion_op(self, dl_framework: DeepLearningFramework):
        if dl_framework == DeepLearningFramework.PYTORCH:
            conversion_op = self.torch_conversion_op
        elif dl_framework == DeepLearningFramework.TENSORFLOW:
            conversion_op = self.tensorflow_conversion_op
        else:
            conversion_op = self.onnx_conversion_op

        return conversion_op

    def execute(
        self,
        model: Any,
        input_data: Union[Iterable, Sequence, DataManager],
        metric_drop_ths: float = None,
        metric: Union[str, Callable] = None,
        optimization_time: str = "constrained",
        dynamic_info: Dict = None,
        config_file: str = None,
        ignore_compilers: List[str] = None,
        ignore_compressors: List[str] = None,
        store_latencies: bool = False,
        **kwargs,
    ):
        self.logger.info(f"Running Speedster on {self.device.name}")

        check_dependencies(self.device)

        if self.fetch_model_op.get_model() is None:
            self.fetch_model_op.execute(model)
        if self.fetch_data_op.get_data() is None:
            self.fetch_data_op.execute(input_data)

        ignore_compilers = map_compilers_and_compressors(
            ignore_compilers, ModelCompiler
        )
        ignore_compressors = map_compilers_and_compressors(
            ignore_compressors, ModelCompressor
        )

        optimization_time = OptimizationTime(optimization_time)

        # Check availability of model and data
        if self.fetch_model_op.get_model() and self.fetch_data_op.get_data():
            self.model = self.fetch_model_op.get_model()
            self.data = self.fetch_data_op.get_data()

            if isinstance(self.data, (DataLoader, tf.data.Dataset)):
                try:
                    self.data = DataManager.from_dataloader(self.data)
                except Exception:
                    raise ValueError(
                        "The provided dataloader does not match the expected "
                        "format.\n"
                        "Speedster supports dataloaders that return tuples in "
                        "the\n"
                        "following formats: \n"
                        "Single input: (input,  label)\n"
                        "Multiple inputs: ((input1, input2, ...),  label) or "
                        "(input1, input2, ...,  label)\n"
                        "Inputs and labels should be either tensors or numpy "
                        "arrays,\n"
                        "depending on the framework used.\n"
                    )

            needs_conversion_to_hf = False
            if is_huggingface_data(self.data[0]):
                (
                    self.model,
                    self.data,
                    input_names,
                    output_structure,
                    output_type,
                ) = convert_hf_model(
                    self.model, self.data, self.device, **kwargs
                )
                needs_conversion_to_hf = True

                if dynamic_info is None:
                    self.logger.warning(
                        "Dynamic shape info has not been provided for the "
                        "HuggingFace model. The resulting optimized model "
                        "will be usable only with a fixed input shape. "
                        "To optimize the model for dynamic shapes, please "
                        "look here: https://nebuly.gitbook.io/nebuly/speeds"
                        "ter/get-started/examples-of-api-options#using-dynamic"
                        "-shape."
                    )

            if not isinstance(self.data, DataManager):
                if check_input_data(self.data):
                    if is_data_subscriptable(self.data):
                        self.data = DataManager(self.data)
                    else:
                        self.data = DataManager.from_iterable(self.data)
                else:
                    raise ValueError(
                        "The provided data does not match the expected "
                        "format.\n"
                        "Speedster supports data in the following formats: \n"
                        "- PyTorch DataLoader\n"
                        "- TensorFlow Dataset\n"
                        "- List of tuples: [((input_0, ... ), label), ...] \n"
                        "Inputs and labels should be either tensors or numpy "
                        "arrays,\n"
                        "depending on the framework used.\n"
                    )

            dl_framework = get_dl_framework(self.model)

            if metric_drop_ths is not None and metric_drop_ths <= 0:
                metric_drop_ths = None
            elif metric_drop_ths is not None and metric is None:
                metric = "numeric_precision"
            if isinstance(metric, str):
                metric = QUANTIZATION_METRIC_MAP.get(metric)

            model_params = extract_info_from_data(
                model=self.model,
                input_data=self.data,
                dl_framework=dl_framework,
                dynamic_info=dynamic_info,
                device=self.device,
            )

            self.data.split(TRAIN_TEST_SPLIT_RATIO)

            model_name = get_model_name(self.model)
            model_info = {
                "model_name": model_name,
                "model_size": read_model_size(model),
                "framework": dl_framework.value,
            }
            self.feedback_collector.store_info(
                key="model_id", value=generate_model_id(model_name)
            )
            self.feedback_collector.store_info(
                key="model_metadata", value=model_info
            )
            self.feedback_collector.store_info(
                key="hardware_setup", value=get_hw_info(self.device)
            )

            # Benchmark original model
            self.orig_latency_measure_op.to(self.device).execute(
                model=self.model,
                input_data=self.data.get_split("test"),
                dl_framework=dl_framework,
            )

            # Store original model result
            original_model_dict = {
                "compiler": dl_framework.value,
                "technique": "original",
                "latency": self.orig_latency_measure_op.get_result()[1],
            }

            with TemporaryDirectory() as tmp_dir:
                tmp_dir = Path(tmp_dir) / "fp32"
                tmp_dir.mkdir(parents=True, exist_ok=True)

                # Convert model to all available frameworks
                self.conversion_op = self._get_conversion_op(dl_framework)
                self.conversion_op.to(self.device).set_state(
                    self.model, self.data
                ).execute(
                    save_path=tmp_dir,
                    model_params=model_params,
                )

                optimized_models = []
                if self.conversion_op.get_result() is not None:
                    original_model_size = (
                        os.path.getsize(self.conversion_op.get_result()[0])
                        if isinstance(self.conversion_op.get_result()[0], str)
                        else len(
                            pickle.dumps(
                                self.conversion_op.get_result()[0], -1
                            )
                        )
                    )
                    for model in self.conversion_op.get_result():
                        optimized_models += self._optimize(
                            model=model,
                            optimization_time=optimization_time,
                            metric_drop_ths=metric_drop_ths,
                            metric=metric,
                            model_params=model_params,
                            ignore_compilers=ignore_compilers,
                            ignore_compressors=ignore_compressors,
                            source_dl_framework=dl_framework,
                        )

            optimized_models.sort(key=lambda x: x[1], reverse=False)

            if len(optimized_models) < 1 or optimized_models[0][0] is None:
                self.logger.warning(
                    "No optimized model has been created. This is likely "
                    "due to a bug in Speedster. Please open an issue and "
                    "report in details your use case."
                )
            else:
                opt_metric_drop = (
                    f"{optimized_models[0][2]:.4f}"
                    if optimized_models[0][2] > MIN_NUMBER
                    else "0"
                )

                metric_name = (
                    "compute_relative_difference"
                    if metric is None
                    else metric.__name__
                )

                orig_latency = self.orig_latency_measure_op.get_result()[1]

                optimizations = self.feedback_collector.get("optimizations")
                optimizations.insert(0, original_model_dict)
                self.feedback_collector.send_feedback()
                if store_latencies:
                    model_id = self.feedback_collector.get("model_id", "")
                    with open(
                        f"{model_name}_latencies_{model_id[:10]}.json", "w"
                    ) as f:
                        json.dump(
                            {
                                "optimizations": optimizations,
                            },
                            f,
                        )
                self.feedback_collector.reset("optimizations")
                self.feedback_collector.reset("model_id")
                self.feedback_collector.reset("model_metadata")

                table = [
                    [
                        "backend",
                        dl_framework.name,
                        optimized_models[0][0].name,
                    ],
                    [
                        "latency",
                        f"{orig_latency:.4f} sec/batch",
                        f"{optimized_models[0][1]:.4f} sec/batch",
                    ],
                    [
                        "throughput",
                        f"{(1 / orig_latency) * model_params.batch_size:.2f} "
                        f"data/sec",
                        f"{1 / optimized_models[0][1]:.2f} data/sec",
                    ],
                    [
                        "model size",
                        f"{original_model_size / 1e6:.2f} MB",
                        f"{optimized_models[0][0].get_size() / 1e6:.2f} MB",
                    ],
                    [f"metric drop ({metric_name})", "", opt_metric_drop],
                    [
                        "speedup",
                        "",
                        f"{orig_latency / optimized_models[0][1]:.2f}x",
                    ],
                ]
                headers = ["Metric", "Original Model", "Optimized Model"]

                self.logger.info(
                    (
                        f"\n[ Speedster results on {self.device.name}]\n"
                        f"{tabulate(table, headers, tablefmt='heavy_outline')}"
                    )
                )

                if needs_conversion_to_hf:
                    from nebullvm.operations.inference_learners.huggingface import (  # noqa: E501
                        HuggingFaceInferenceLearner,
                    )

                    self.optimal_model = HuggingFaceInferenceLearner(
                        core_inference_learner=optimized_models[0][0],
                        output_structure=output_structure,
                        input_names=input_names,
                        output_type=output_type,
                    )
                else:
                    self.optimal_model = optimized_models[0][0]

    def _optimize(
        self,
        model: Any,
        optimization_time: OptimizationTime,
        metric_drop_ths: float,
        metric: Callable,
        model_params: ModelParams,
        ignore_compilers: List[ModelCompiler],
        ignore_compressors: List[ModelCompressor],
        source_dl_framework: DeepLearningFramework,
    ) -> List[BaseInferenceLearner]:
        if self.orig_latency_measure_op.get_result() is not None:
            model_outputs = self.orig_latency_measure_op.get_result()[0]
            if isinstance(model, Module):
                optimization_op = self.torch_optimization_op
            elif isinstance(model, tf.Module) and model is not None:
                optimization_op = self.tensorflow_optimization_op
            else:
                optimization_op = self.onnx_optimization_op

            optimization_op.to(self.device).execute(
                model=model,
                input_data=self.data,
                optimization_time=optimization_time,
                metric_drop_ths=metric_drop_ths,
                metric=metric,
                model_params=model_params,
                model_outputs=model_outputs,
                ignore_compilers=ignore_compilers,
                ignore_compressors=ignore_compressors,
                source_dl_framework=source_dl_framework,
            )

            optimized_models = optimization_op.get_result()

            return optimized_models

    def get_result(self) -> Any:
        return self.optimal_model
